;BMP085.H --- BMP085 Barometer Include File
;Thomas Henry --- 8/11/2014

;There are three commands that the calling program will use:
;
;getConstants()
;getTemp(Celsius, Fahrenheit)
;getPressure(Pascals, Station, MSLP)

;The calling program may also set the following constants.
;See the descriptions below for their meaning and defaults.
;
;BMP_PRESS_FACTOR
;BMP_FUDGE_FACTOR1
;BMP_FUDGE_FACTOR2
;BMP_OSS

;----- Constants

#define BMP_ADDR  0xEE              ;BMP085 device address

;The following constant is used to convert from station pressure
;to mean sea level pressure. It depends upon your altitude in
;meters. Compute the rounded value like this:
;(1/((1-altitude/44330)^5.255))*10000

#ifndef BMP_PRESS_FACTOR
  #define BMP_PRESS_FACTOR  10000   ;defaults to sea level
#endif

;And the next two can be used to diddle the pressure to match
;an existing nearby barometer, else just use the defaults
;of zero. The numbers are in hundredths of an inch of mercury.

#ifndef BMP_FUDGE_FACTOR1
  #define BMP_FUDGE_FACTOR1 0
#endif

#ifndef BMP_FUDGE_FACTOR2
  #define BMP_FUDGE_FACTOR2 0
#endif

#ifndef BMP_OSS
  #define BMP_OSS 3                 ;default to mode 3
#endif

;The following three constants are Computed at compile time.
;BMP_MODE is the command number for the sampling mode chosen.
;BMP_DELAY is the time in mS that particular mode requires.
;BMP_OSSDIV is the right shift amount for oversampling.
;All depend upon the mode selected by BMP_OSS (0, 1, 2 or 3).

#script
  if BMP_OSS = 0 then
    BMP_MODE = 0x34                 ;command number
    BMP_DELAY = 5                   ;conversion time in mS
    BMP_OSSDIV = 8                  ;oversampling divisor
  end if

  if BMP_OSS = 1 then
    BMP_MODE = 0x74
    BMP_DELAY = 8
    BMP_OSSDIV = 7
  end if

  if BMP_OSS = 2 then
    BMP_MODE = 0xB4
    BMP_DELAY = 14
    BMP_OSSDIV = 6
  end if

  if BMP_OSS = 3 then
    BMP_MODE = 0xF4
    BMP_DELAY = 26
    BMP_OSSDIV = 5
  end if
#endscript

;----- Variables

;Here are the calibration values read from the BMP085.
;Note that bmp_ac4, bmp_ac5 and bmp_ac6 are unsigned and the
;remainder are signed (2's-complement). Further
;observe that most values throughout are promoted
;from word or integer types (16 bits) to long (32 bits).

dim bmp_ac1, bmp_ac2, bmp_ac3, bmp_ac4, bmp_ac5, bmp_ac6 as long
dim bmp_b1, bmp_b2, bmp_mb, bmp_mc, bmp_md as long

;The next variables hold the uncompensated temperature
;and pressure readings from the BMP085.

dim bmp_ut, bmp_up as long

;And these are the subroutine variables.

dim bmp_copy1, bmp_copy2, bmp_temp as long
dim bmp_b3, bmp_b4, bmp_b5, bmp_b6, bmp_b7, bmp_x1, bmp_x2, bmp_x3 as long
dim bmp_MSB, bmp_LSB, bmp_XLSB, as byte

;----- High Level Subroutines

sub getConstants()
  I2CStart
  I2CSend(BMP_ADDR)                 ;wake up the device
  I2CSend(0xAA)                     ;address first calibration constant
  I2CStop

  I2CStart
  I2CSend(BMP_ADDR+1)               ;indicate a read operation

  ;The values within the BMP085 are either signed or unsigned
  ;16-bit numbers. However, they are all promoted to signed
  ;32-bit numbers for use within this program.

  bmp_ac1 = BMPreadSigned()         ;get and promote to longs
  bmp_ac2 = BMPreadSigned()
  bmp_ac3 = BMPreadSigned()
  bmp_ac4 = BMPreadUnsigned()
  bmp_ac5 = BMPreadUnsigned()
  bmp_ac6 = BMPreadUnsigned()
  bmp_b1  = BMPreadSigned()
  bmp_b2  = BMPreadSigned()
  bmp_mb  = BMPreadSigned()
  bmp_mc  = BMPreadSigned()
  bmp_md  = BMPreadSigned()
  I2CReceive(bmp_MSB, NACK)         ;dummy read means end
  I2CStop
end sub

;-----

sub getTemp(out BMPCels as integer, out BMPFahr as integer)
  ;Fetch the current temperature.

  BMPreadI2C(0x2E, 5, bmp_MSB, bmp_LSB, bmp_XLSB)
  ;bmp_XLSB is ignored here
  bmp_ut = BMPcastSigned(bmp_MSB,bmp_LSB)

  ;The following conversion is straight from the data sheet. However
  ;much care must be taken with regard to the order of operations,
  ;data types and rounding errors.

  bmp_x1 = bmp_ut - bmp_ac6
  bmp_x1 = BMPsignedLongMult(bmp_x1,bmp_ac5)
  bmp_x1 = BMPshiftR(bmp_x1,15)

  bmp_temp = bmp_x1 + bmp_md
  bmp_x2 = BMPshiftL(bmp_mc,11)
  bmp_x2 = BMProundedDiv(bmp_x2,bmp_temp)

  bmp_b5 = bmp_x1 + bmp_x2

  bmp_temp = bmp_b5 + 8
  bmp_temp = BMPshiftR(bmp_temp,4)

  BMPCels = bmp_temp                ;cast Celsius to integer
  BMPFahr = (18*BMPCels+3205)/10    ;also return Fahrenheit
end sub

;-----

sub getPressure(out BMPPascals as long, out BMPStation as integer, out BMPMslp as integer)
  ;Fetch the current pressure.

  BMPreadI2C(BMP_MODE, BMP_DELAY, bmp_MSB, bmp_LSB, bmp_XLSB)
  bmp_up = bmp_XLSB                 ;read the three bytes
  bmp_up_h = bmp_LSB                ;which make up the
  bmp_up_u = bmp_MSB                ;current measurement
  bmp_up_e = 0
  bmp_up = BMPshiftR(bmp_up,BMP_OSSDIV) ;a 16-, 17-, 18- or 19-bit result

  ;The following conversion is straight from the data sheet. However
  ;much care must be taken with regard to the order of operations,
  ;data types and rounding errors. Most of the computations involve
  ;signed 32-bit numbers.

  bmp_b6 = bmp_b5 - 4000

  bmp_x1 = BMPsignedLongMult(bmp_b6,bmp_b6)
  bmp_x1 = BMPshiftR(bmp_x1,12)
  bmp_x1 = BMPsignedLongMult(bmp_b2,bmp_x1)
  bmp_x1 = BMPshiftR(bmp_x1,11)

  bmp_x2 = BMPsignedLongMult(bmp_ac2,bmp_b6)
  bmp_x2 = BMPshiftR(bmp_x2,11)

  bmp_x3 = bmp_x1 + bmp_x2

  bmp_b3 = BMPsignedLongMult(bmp_ac1,4) + bmp_x3
  bmp_b3 = BMPshiftL(bmp_b3,BMP_OSS) + 2
  bmp_b3 = BMPshiftR(bmp_b3,2)

  bmp_x1 = BMPsignedLongMult(bmp_ac3,bmp_b6)
  bmp_x1 = BMPshiftR(bmp_x1,13)

  bmp_x2 = BMPsignedLongMult(bmp_b6,bmp_b6)
  bmp_temp = BMPshiftR(bmp_x2,12)
  bmp_x2 = BMPsignedLongMult(bmp_b1,bmp_temp)
  bmp_x2 = BMPshiftR(bmp_x2,16)

  bmp_x3 = bmp_x1 + bmp_x2 + 2
  bmp_x3 = BMPshiftR(bmp_x3,2)

  bmp_b4 = bmp_x3 + 32768
  bmp_b4 = BMPsignedLongMult(bmp_ac4,bmp_b4)
  bmp_b4 = BMPshiftR(bmp_b4,15)

  bmp_b7 = bmp_up-bmp_b3
  bmp_temp = BMPshiftR(50000,BMP_OSS)
  bmp_b7 = BMPsignedLongMult(bmp_b7,bmp_temp)

  if bmp_b7 < 0x80000000 then
    BMPPascals = BMPshiftL(bmp_b7,1)
    BMPPascals = BMPPascals / bmp_b4  ;unsigned, unrounded division
  else
    BMPPascals = bmp_b7 / bmp_b4      ;unsigned, unrounded division
    BMPPascals = BMPshiftL(BMPPascals,1)
  end if

  ;The next four lines carry out squaring to two additional
  ;decimal BMPplaces, concluding by rounding to the nearest whole number.

  bmp_x1 = BMPPascals * 100           ;unsigned multiplication
  bmp_x1 = bmp_x1 / 256               ;unsigned, unrounded division
  bmp_x1 = bmp_x1 * bmp_x1 + 50000    ;ditto for this and the next
  bmp_x1 = bmp_x1 / 10000

  bmp_x1 = BMPsignedLongMult(bmp_x1,3038)
  bmp_x1 = BMPshiftR(bmp_x1,16)

  ;0xFFFFE343 is -7357 in 2's-complement form
  bmp_x2 = BMPsignedLongMult(0xFFFFE343,BMPPascals)
  bmp_x2 = BMPsignedLongDiv(bmp_x2,65536)

  bmp_temp = bmp_x1 + bmp_x2 + 3791
  bmp_temp = BMProundedDiv(bmp_temp,16) ;signed, rounded division
  BMPPascals += bmp_temp                ;station pressure in Pascals

  ;Now convert to get station pressure in inHg, rounded to hundredth

  bmp_temp = (BMPPascals * 10000) / 33864 + 5
  bmp_temp = bmp_temp / 10
  BMPStation = bmp_temp + BMP_FUDGE_FACTOR1

  ;Finally convert to get mean sea level pressure in inHg

  bmp_temp = [long]BMPPascals * BMP_PRESS_FACTOR
  bmp_temp = (bmp_temp / 33864 + 5) / 10
  BMPMslp = bmp_temp + BMP_FUDGE_FACTOR2
end sub

;----- Low Level Subroutines

function BMPreadSigned() as long      ;read a signed value
  I2CReceive(bmp_MSB)                 ;the MSB is stored first
  I2CReceive(bmp_LSB)                 ;followed by the LSB next
  BMPreadSigned = BMPcastSigned(bmp_MSB, bmp_LSB)
end function

;-----

function BMPreadUnsigned() as long    ;read an unsigned number
  I2CReceive(bmp_MSB)                 ;the MSB is stored first
  I2CReceive(bmp_LSB)                 ;followed by the LSB
  BMPreadUnsigned = BMPcastUnsigned(bmp_MSB, bmp_LSB)
end function

;-----

function BMPcastSigned(msb as byte, lsb as byte) as long
  ;Convert a signed 16-bit number to signed 32-bit number.

  BMPcastSigned = lsb
  BMPcastSigned_h = msb
  if msb.7 then
    BMPcastSigned_u = 0xFF            ;negative number
    BMPcastSigned_e = 0xFF
  else
    BMPcastSigned_u = 0x00            ;non-negative number
    BMPcastSigned_e = 0x00
  end if
end function

;-----

function BMPcastUnsigned(msb as byte, lsb as byte) as long
  ;Convert an unsigned 16-bit number to signed 32-bit number.

  BMPcastUnsigned = lsb
  BMPcastUnsigned_h = msb
  BMPcastUnsigned_u = 0x00            ;always non-negative
  BMPcastUnsigned_e = 0x00
end function

;-----

sub BMPreadI2C(in BMPcmd, in BMPwaitTime, out BMPval1, out BMPval2, out BMPval3)
  ;Read uncompensated temperature or pressure from the BMP085.

  I2CStart
  I2CSend(BMP_ADDR)                   ;address for a write
  I2CSend(0xF4)                       ;address the control register
  I2CSend(BMPcmd)                     ;give the command
  I2CStop

  wait BMPwaitTime mS                 ;conversion time

  I2CStart
  I2CSend(BMP_ADDR)                   ;address for a write
  I2CSend(0xF6)                       ;set the address
  I2CStop

  I2CStart
  I2CSend(BMP_ADDR+1)                 ;address for a read
  I2CReceive(BMPval1)                 ;get three consecbmp_utive values
  I2CReceive(BMPval2)
  I2CReceive(BMPval3, NACK)           ;last is meaningless for temperature
  I2CStop
end sub

;-----

function BMPshiftL(BMPparam as long, BMPplaces as byte) as long
  ;Left-shift a long number.

  repeat BMPplaces
    status.c = 0
    rotate BMPparam left
  end repeat
  BMPshiftL = BMPparam
end function

;-----

function BMPshiftR(BMPparam as long, BMPplaces as byte) as long
  ;Right-shift a long number.

  repeat BMPplaces
    status.c = 0
    rotate BMPparam right
  end repeat
  BMPshiftR = BMPparam
end function

;-----

function BMPsignedLongMult(BMPop1 as long, BMPop2 as long) as long
  ;Multiply two signed long numbers.

  bmp_copy1 = BMPop1
  bmp_copy2 = BMPop2

  if BMPop1.31 then
    bmp_copy1 = (not BMPop1) + 1      ;take absolute value
  end if

  if BMPop2.31 then
    bmp_copy2 = (not BMPop2) + 1      ;likewise here
  end if

  BMPsignedLongMult = bmp_copy1 * bmp_copy2

  bmp_MSB = BMPop1_e XOR BMPop2_e
  if bmp_MSB.7 then                   ;mixed signs mean negative
    BMPsignedLongMult = (not BMPsignedLongMult)+1
  end if
end function

;-----

function BMPsignedLongDiv(BMPop1 as long, BMPop2 as long) as long
  ;Divide two signed long numbers with truncation.

  bmp_copy1 = BMPop1
  bmp_copy2 = BMPop2

  if BMPop1.31 then
    bmp_copy1 = (not BMPop1) + 1      ;take absolute value
  end if

  if BMPop2.31 then
    bmp_copy2 = (not BMPop2) + 1      ;likewise here
  end if

  BMPsignedLongDiv = bmp_copy1 / bmp_copy2

  bmp_MSB = BMPop1_e XOR BMPop2_e
  if bmp_MSB.7 then                   ;mixed signs mean negative
    BMPsignedLongDiv = (not BMPsignedLongDiv) + 1
  end if
end function

;-----

function BMProundedDiv(BMPop1 as long, BMPop2 as long) as long
  ;Divide two signed long numbers, rounded to nearest whole number.

  bmp_copy1 = BMPop1
  bmp_copy2 = BMPop2

  if BMPop1.31 then
    bmp_copy1 = (not BMPop1) + 1      ;take absolute value
  end if

  if BMPop2.31 then
    bmp_copy2 = (not BMPop2) + 1      ;likewise here
  end if

  bmp_copy1 = bmp_copy1 * 10          ;scale bmp_up first

  BMProundedDiv = bmp_copy1 / bmp_copy2
  BMProundedDiv = BMProundedDiv + 5   ;then round result
  BMProundedDiv = BMProundedDiv / 10

  bmp_MSB = BMPop1_e XOR BMPop2_e
  if bmp_MSB.7 then                   ;mixed signs mean negative
    BMProundedDiv = (not BMProundedDiv)+1
  end if
end function

